﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Windows.Forms;

namespace Paketbote
{
    class Gameboard
    {
        
        //orientations of a package
        public struct Package_Points
        {
            public Package_Points_Orientation zero;
            public Package_Points_Orientation one;
            public Package_Points_Orientation two;
            public Package_Points_Orientation three;
        }
        //slots to be evaluated and updated on rotation
        public struct Package_Points_Orientation
        {
            public List<Point> check;
            public List<Point> delete;
        }

        public static Package_Points pkg_T = new Package_Points();
        public static Package_Points pkg_L = new Package_Points();
        public static Package_Points pkg_J = new Package_Points();
        public static Package_Points pkg_Z = new Package_Points();
        public static Package_Points pkg_S = new Package_Points();
        public static Package_Points pkg_I = new Package_Points();
        public static Package_Points_Orientation pkg_current = new Package_Points_Orientation();


        public Package[,] matrix;
        public int score { get; set; }
        public int rows { get; set; }
        public int cols { get; set; }
        public bool stop_Key_Input { get; set; }
        public Point top { get; set; }
        private Pen pen_black;

        //points to be redrawn
        public List<Point> update_points;

        /// <summary>
        /// spawns the package in the spawn-area of the gameboard
        /// </summary>
        /// <param name="type">type of package</param>
        /// <param name="col">col of center package</param>
        public void spawn_Package(enum_type type, int col)
        {
            List<Point> spawn_Points = get_spawn_Points(type);
            matrix[4, col].type = type;
            matrix[4, col].center = true;
            matrix[4,col].state = enum_state.moving;         
            matrix[4,col].orientation = enum_orientation.zero;
            matrix[4, col].points = Package.get_Points(type);

            for (int i = 0; i < spawn_Points.Count; i++)
            {
                matrix[4 + spawn_Points[i].Y, col + spawn_Points[i].X].type = type;
                matrix[4 + spawn_Points[i].Y, col + spawn_Points[i].X].state = enum_state.moving;
                matrix[4 + spawn_Points[i].Y, col + spawn_Points[i].X].points = Package.get_Points(type);
            }
        }

        /// <summary>
        /// gets the spawn points for a whole package around the center package
        /// </summary>
        /// <param name="type">type of package</param>
        /// <returns>points around the center package</returns>
        public List<Point> get_spawn_Points(enum_type type)
        {
            List<Point> spawn_points;
            switch (type)
            {
                case enum_type.T:
                    spawn_points = new List<Point> {new Point(-1,0),new Point(0,-1),new Point(1,0)};
                    break;
                case enum_type.L:
                    spawn_points = new List<Point> { new Point(-2,0), new Point(-1,0), new Point(0,-1)};
                    break;
                case enum_type.J:
                    spawn_points = new List<Point> { new Point(2,0), new Point(1,0), new Point(0,-1)};
                    break;
                case enum_type.Z:
                    spawn_points = new List<Point> { new Point(-1,-1), new Point(0,-1), new Point(1,0)};
                    break;
                case enum_type.S:
                    spawn_points = new List<Point> { new Point(1,-1), new Point(0,-1), new Point(-1,0)};
                    break;
                case enum_type.I:
                    spawn_points = new List<Point> { new Point(0,-1), new Point(0,-2), new Point(0,-3)};
                    break;
                case enum_type.O:
                    spawn_points = new List<Point> { new Point(0,-1), new Point(1,-1), new Point(1,0)};
                    break;
                default:
                    spawn_points = new List<Point> { };
                    break;
            }
            return spawn_points;
        }

        //returns next orientation state
        public enum_orientation get_next_Orientation(enum_orientation orientation,string direction = "right")
        {
            if (direction=="right")
            {
                switch (orientation)
                {
                    case enum_orientation.zero:
                        return enum_orientation.one;
                    case enum_orientation.one:
                        return enum_orientation.two;
                    case enum_orientation.two:
                        return enum_orientation.three;
                    case enum_orientation.three:
                        return enum_orientation.zero;
                    default:
                        return enum_orientation.zero;
                }
            }
            else
            {
                switch (orientation)
                {
                    case enum_orientation.zero:
                        return enum_orientation.three;
                    case enum_orientation.one:
                        return enum_orientation.zero;
                    case enum_orientation.two:
                        return enum_orientation.one;
                    case enum_orientation.three:
                        return enum_orientation.two;
                    default:
                        return enum_orientation.zero;
                }
            }

        }

        //returns the corresponding package
        public Package_Points_Orientation Get_Points_Orientation(enum_type type, enum_orientation orientation)
        {
            switch (type)
            {
                case enum_type.T:
                    switch (orientation)
                    {
                        case enum_orientation.zero:
                            return pkg_T.zero;
                        case enum_orientation.one:
                            return pkg_T.one;
                        case enum_orientation.two:
                            return pkg_T.two;
                        case enum_orientation.three:
                            return pkg_T.three;
                        default: return new Package_Points_Orientation();
                    }
                case enum_type.L:
                    switch (orientation)
                    {
                        case enum_orientation.zero:
                            return pkg_L.zero;
                        case enum_orientation.one:
                            return pkg_L.one;
                        case enum_orientation.two:
                            return pkg_L.two;
                        case enum_orientation.three:
                            return pkg_L.three;
                        default: return new Package_Points_Orientation();
                    }
                case enum_type.J:
                    switch (orientation)
                    {
                        case enum_orientation.zero:
                            return pkg_J.zero;
                        case enum_orientation.one:
                            return pkg_J.one;
                        case enum_orientation.two:
                            return pkg_J.two;
                        case enum_orientation.three:
                            return pkg_J.three;
                        default: return new Package_Points_Orientation();
                    }
                case enum_type.Z:
                    switch (orientation)
                    {
                        case enum_orientation.zero:
                            return pkg_Z.zero;
                        case enum_orientation.one:
                            return pkg_Z.one;
                        case enum_orientation.two:
                            return pkg_Z.two;
                        case enum_orientation.three:
                            return pkg_Z.three;
                        default: return new Package_Points_Orientation();
                    }
                case enum_type.S:
                    switch (orientation)
                    {
                        case enum_orientation.zero:
                            return pkg_S.zero;
                        case enum_orientation.one:
                            return pkg_S.one;
                        case enum_orientation.two:
                            return pkg_S.two;
                        case enum_orientation.three:
                            return pkg_S.three;
                        default: return new Package_Points_Orientation();
                    }
                case enum_type.I:
                    switch (orientation)
                    {
                        case enum_orientation.zero:
                            return pkg_I.zero;
                        case enum_orientation.one:
                            return pkg_I.one;
                        case enum_orientation.two:
                            return pkg_I.two;
                        case enum_orientation.three:
                            return pkg_I.three;
                        default: return new Package_Points_Orientation();
                    }
                case enum_type.O:
                    Package_Points_Orientation pkg_O;
                    pkg_O.check = new List<Point> { };
                    pkg_O.delete = new List<Point> { };
                    return pkg_O; //return empty lists so no packages will be updated on rotation
                default: return new Package_Points_Orientation();
            }
        }

        /// <summary>
        /// creates a new instance of a gameboard
        /// </summary>
        /// <param name="rows">number of visible rows</param>
        /// <param name="cols">number of visible columns</param
        /// <param name="top">top left position of gameboard</param>
        public Gameboard(int rows, int cols, Point top)
        {
            //definition of all the 6 packages regarding orientation
            #region pkg_T
            pkg_T.zero.check = new List<Point> {new Point(0,1)};
            pkg_T.zero.delete = new List<Point> {new Point(-1,0)};
            pkg_T.one.check = new List<Point> {new Point(-1,0)};
            pkg_T.one.delete = new List<Point> {new Point(0,-1)};
            pkg_T.two.check = new List<Point> {new Point(0,-1)};
            pkg_T.two.delete = new List<Point> {new Point(1,0)};
            pkg_T.three.check = new List<Point> {new Point(1,0)};
            pkg_T.three.delete = new List<Point> {new Point(0,1)};
            #endregion
            #region pkg_L
            pkg_L.zero.check = new List<Point> {new Point(0,-2),new Point(1,0)};
            pkg_L.zero.delete = new List<Point> {new Point(-2,0),new Point(-1,0)};
            pkg_L.one.check = new List<Point> {new Point(2,0),new Point(0,1)};
            pkg_L.one.delete = new List<Point> {new Point(0,-2),new Point(0,-1)};
            pkg_L.two.check = new List<Point> {new Point(0,2),new Point(-1,0)};
            pkg_L.two.delete = new List<Point> {new Point(2,0),new Point(1,0)};
            pkg_L.three.check = new List<Point> {new Point(-2,0),new Point(0,-1)};
            pkg_L.three.delete = new List<Point> {new Point(0,2),new Point(0,1)};
            #endregion
            #region pkg_J
            pkg_J.zero.check = new List<Point> {new Point(0,2),new Point(0,1)};
            pkg_J.zero.delete = new List<Point> {new Point(0,-1),new Point(2,0)};
            pkg_J.one.check = new List<Point> {new Point(-2,0),new Point(-1,0)};
            pkg_J.one.delete = new List<Point> {new Point(1,0),new Point(0,2)};
            pkg_J.two.check = new List<Point> {new Point(0,-2),new Point(0,-1)};
            pkg_J.two.delete = new List<Point> {new Point(0,1),new Point(-2,0)};
            pkg_J.three.check = new List<Point> {new Point(2,0),new Point(1,0)};
            pkg_J.three.delete = new List<Point> {new Point(-1,0),new Point(0,-2)};
            #endregion pkg_J
            #region pkg_Z
            pkg_Z.zero.check = new List<Point> {new Point(1, -1),new Point(0,1)};
            pkg_Z.zero.delete = new List<Point> {new Point(0, -1),new Point(-1,-1)};
            pkg_Z.one.check = new List<Point> {new Point(-1,0),new Point(1,1)};
            pkg_Z.one.delete = new List<Point> {new Point(1, 0),new Point(1,-1)};
            pkg_Z.two.check = new List<Point> {new Point(0,-1),new Point(-1,1)};
            pkg_Z.two.delete = new List<Point> {new Point(0,1),new Point(1,1)};
            pkg_Z.three.check = new List<Point> {new Point(-1, -1),new Point(1,0)};
            pkg_Z.three.delete = new List<Point> {new Point(-1, 0),new Point(-1,1)};
            #endregion
            #region pkg_S
            pkg_S.zero.check = new List<Point> {new Point(1,0),new Point(1,1)};
            pkg_S.zero.delete = new List<Point> {new Point(-1,0),new Point(1,-1)};
            pkg_S.one.check = new List<Point> {new Point(0,1),new Point(-1,1)};
            pkg_S.one.delete = new List<Point> {new Point(0,-1),new Point(1,1)};
            pkg_S.two.check = new List<Point> {new Point(-1,0),new Point(-1,-1)};
            pkg_S.two.delete = new List<Point> {new Point(1,0),new Point(-1,1)};
            pkg_S.three.check = new List<Point> {new Point(0,-1),new Point(1,-1)};
            pkg_S.three.delete = new List<Point> {new Point(0,1),new Point(-1,-1)};
            #endregion
            #region pkg_I
            pkg_I.zero.check = new List<Point> {new Point(1,0),new Point(2,0),new Point(3,0)};
            pkg_I.zero.delete = new List<Point> {new Point(0,-1),new Point(0,-2),new Point(0,-3)};
            pkg_I.one.check = new List<Point> {new Point(0,1),new Point(0,2),new Point(0,3)};
            pkg_I.one.delete = new List<Point> {new Point(1,0),new Point(2,0),new Point(3,0)};
            pkg_I.two.check = new List<Point> {new Point(-1,0),new Point(-2,0),new Point(-3,0)};
            pkg_I.two.delete = new List<Point> {new Point(0,1),new Point(0,2),new Point(0,3)};
            pkg_I.three.check = new List<Point> {new Point(0,-1),new Point(0,-2),new Point(0,-3)};
            pkg_I.three.delete = new List<Point> {new Point(-1,0),new Point(-2,0),new Point(-3,0)};
            #endregion
            //make the matrix 6 rows larger; 5 for spawn and 1 for ground
            this.rows = rows+6;
            this.cols = cols;
            matrix = new Package[this.rows, this.cols];
            stop_Key_Input = false;
            this.top = top;
            update_points = new List<Point> { };
            //initialize matrix
            for (int row = 0; row < this.rows; row++)
            {
                for (int col = 0; col < this.cols; col++)
                {
                    if (row==this.rows-1)
                    {
                        matrix[row, col] = new Package(enum_type.ground, enum_state.placed, enum_orientation.zero, 0, false);
                    }
                    else
                    {
                        matrix[row, col] = new Package();
                    }
                    
                }
            }
            pen_black = new Pen(Color.Black, 5);
            pen_black.Alignment = System.Drawing.Drawing2D.PenAlignment.Inset;
        }
        /// <summary>
        /// draws border around the gameboard
        /// </summary>
        public void draw_Border()
        {
            frm_Paketbote.form_Graphics.DrawRectangle(pen_black, top.X-5, top.Y-5, cols * 40 + 10, (rows - 6) * 40 + 10);
        }

        /// <summary>
        /// fills the complete gameboard with air-packages
        /// </summary>
        public void clear_Gameboard()
        {
            for (int row = 0; row < this.rows; row++)
            {
                for (int col = 0; col < this.cols; col++)
                {
                    if (row == this.rows - 1)
                    {
                        matrix[row, col] = new Package(enum_type.ground, enum_state.placed, enum_orientation.zero, 0, false);
                    }
                    else
                    {
                        matrix[row, col] = new Package();
                    }
                }
            }
        }

        /// <summary>
        /// calculates the next gamestate depending on the direction of package-movement
        /// </summary>
        /// <param name="direction">direction in which package should be moved</param>
        /// <returns>1 if move was possible, 0 if impossible and 2 if game over</returns>
        public byte calculate_next_Gamestate(string direction)
        {
            switch (direction)
            {
                //scans gameboard ↑ and → to check if each moving package can be moved 1 slot down
                case "down":
                    for (int row = rows - 2; row >= 0; row--)
                    {
                        for (int col = 0; col < cols; col++)
                        {
                            if (row < 5 && row>= 0 && matrix[row, col].type != enum_type.air)
                            {
                                if (matrix[row + 1, col].state == enum_state.placed && matrix[row + 1, col].type != enum_type.air)
                                {
                                    //game over since package can't be moved down and spawn area has still packages
                                    return 2;
                                }
                            }
                            if (matrix[row, col].state == enum_state.moving && matrix[row + 1, col].state == enum_state.placed && matrix[row + 1, col].type != enum_type.air)
                            {
                                return 0;
                            }
                        }
                    }
                    break;
                //scans gameboard → and ↓ to check if each moving package can be moved 1 slot left
                case "left":
                    for (int col = 0; col < cols; col++)
                    {
                        for (int row = 0; row < rows-1; row++)
                        {
                            if (col == 0 && matrix[row, col].state == enum_state.moving)
                            {
                                return 0;
                            }
                            else if (matrix[row, col].state == enum_state.moving && matrix[row, col -1 ].state == enum_state.placed && matrix[row, col -1 ].type != enum_type.air)
                            {
                                return 0;
                            }
                            else { }
                        }
                    }
                    break;
                //scans gameboard → and ↓ to check if each moving package can be moved 1 slot right
                case "right":
                    for (int col = 0; col < cols; col++)
                    {
                        for (int row = 0; row < rows-1; row++)
                        {
                            if (col == cols-1 && matrix[row, col].state == enum_state.moving)
                            {
                                return 0;
                            }
                            else if (matrix[row, col].state == enum_state.moving && matrix[row, col + 1].state == enum_state.placed && matrix[row, col + 1].type != enum_type.air)
                            {
                                return 0;
                            }
                            else { }
                        }
                    }
                    break;
                //scans gameboard → and ↓ to check if package can be rotated to the right
                case "rotate_right":
                    for (int col = 0; col < cols; col++)
                    {
                        for (int row = 0; row < rows-1; row++)
                        {
                            if (matrix[row, col].center == true && matrix[row, col].state == enum_state.moving)
                            {
                                pkg_current = Get_Points_Orientation(matrix[row, col].type, matrix[row, col].orientation);
                                for (int i = 0; i < pkg_current.check.Count; i++)
                                {
                                    if (row + pkg_current.check[i].Y < rows && col + pkg_current.check[i].X < cols && col + pkg_current.check[i].X >= 0)
                                    {
                                        if (matrix[row + pkg_current.check[i].Y, col + pkg_current.check[i].X].state == enum_state.placed && matrix[row + pkg_current.check[i].Y, col + pkg_current.check[i].X].type != enum_type.air)
                                        {
                                            return 0;
                                        }
                                    }
                                    else { return 0; }
                                }
                            }
                        }
                    }
                    break;
                //scans gameboard → and ↓ to check if package can be rotated to the left
                case "rotate_left":
                    for (int col = 0; col < cols; col++)
                    {
                        for (int row = 0; row < rows - 1; row++)
                        {
                            if (matrix[row, col].center == true && matrix[row, col].state == enum_state.moving)
                            {                         
                                pkg_current = Get_Points_Orientation(matrix[row, col].type, get_next_Orientation(matrix[row, col].orientation, "left"));
                                for (int i = 0; i < pkg_current.check.Count; i++)
                                {
                                    if (row + pkg_current.delete[i].Y < rows && col + pkg_current.delete[i].X < cols && col + pkg_current.delete[i].X >= 0)
                                    {
                                        if (matrix[row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X].state == enum_state.placed && matrix[row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X].type != enum_type.air)
                                        {
                                            return 0;
                                        }
                                    }
                                    else { return 0; }
                                }
                            }
                        }
                    }
                    break;
                default:
                    break;
            }
            return 1;
        }
        /// <summary>
        /// updates the position of the package according to the direction
        /// </summary>
        /// <param name="direction">direction in which package was moved</param>
        /// <returns>true if package was placed on bottom</returns>
        public bool update_next_Gamestate(string direction)
        {
            bool flag = false;
            update_points.Clear();

            for (int row = rows - 2; row >= 0; row--)
            {
                switch(direction)
                {
                    //scans gameboard ↑ and → to place each moving package 1 slot down
                    case "down":
                        for (int col = 0; col < cols; col++)
                        {
                            if (matrix[row, col].state == enum_state.moving)
                            {
                                if (row == rows - 2)
                                {
                                    flag = true;
                                }
                                matrix[row + 1, col] = matrix[row, col];
                                matrix[row, col] = new Package();
                                update_points.Add(new Point(row + 1, col));
                                update_points.Add(new Point(row, col));
                            }
                        }
                        break;
                    //scans gameboard ↑ and → to place each moving package 1 slot left
                    case "left":
                        for (int col = 0; col < cols; col++)
                        {
                            if (matrix[row, col].state == enum_state.moving)
                            {
                                matrix[row, col - 1] = matrix[row, col];
                                matrix[row, col] = new Package();
                                update_points.Add(new Point(row, col - 1));
                                update_points.Add(new Point(row, col));
                            }
                        }
                        break;
                    //scans gameboard ↑ and ← to place each moving package 1 slot right
                    case "right":
                        for (int col = cols-1; col >= 0; col--)
                        {
                            if (matrix[row, col].state == enum_state.moving)
                            {
                               matrix[row, col + 1] = matrix[row, col];
                               matrix[row, col] = new Package();
                               update_points.Add(new Point(row, col + 1));
                               update_points.Add(new Point(row, col));
                            }
                        }
                        break;
                    //scans gameboard ↑ and → to rotate the current package right
                    case "rotate_right":
                        for (int col = 0; col < cols; col++)
                        {
                            if (matrix[row, col].center == true && matrix[row, col].state == enum_state.moving)
                            {
                                pkg_current = Get_Points_Orientation(matrix[row, col].type, matrix[row, col].orientation);

                                for (int i = 0; i < pkg_current.delete.Count; i++)
                                {
                                    matrix[row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X] = new Package();
                                    update_points.Add(new Point(row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X));
                                }
                                for (int i = 0; i < pkg_current.check.Count; i++)
                                {
                                    matrix[row + pkg_current.check[i].Y, col + pkg_current.check[i].X].type = matrix[row, col].type;
                                    matrix[row + pkg_current.check[i].Y, col + pkg_current.check[i].X].state = matrix[row, col].state;
                                    matrix[row + pkg_current.check[i].Y, col + pkg_current.check[i].X].points = matrix[row, col].points;
                                    update_points.Add(new Point(row + pkg_current.check[i].Y, col + pkg_current.check[i].X));
                                }
                                matrix[row, col].orientation = get_next_Orientation(matrix[row, col].orientation);
                            }
                        }
                        break;
                        //scans gameboard ↑ and → to rotate the current package left
                        //note that delete and check are in reverse order and also the previous orientation is used instead
                        //That is because rotating left is the same as rotating right on the previous orientation in this implementation
                    case "rotate_left":
                        for (int col = 0; col < cols; col++)
                        {
                            if (matrix[row, col].center == true && matrix[row, col].state == enum_state.moving)
                            {
                                pkg_current = Get_Points_Orientation(matrix[row, col].type, get_next_Orientation(matrix[row, col].orientation, "left"));

                                for (int i = 0; i < pkg_current.check.Count; i++)
                                {
                                    matrix[row + pkg_current.check[i].Y, col + pkg_current.check[i].X] = new Package();
                                    update_points.Add(new Point(row + pkg_current.check[i].Y, col + pkg_current.check[i].X));
                                }
                                for (int i = 0; i < pkg_current.delete.Count; i++)
                                {
                                    matrix[row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X].type = matrix[row, col].type;
                                    matrix[row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X].state = matrix[row, col].state;
                                    matrix[row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X].points = matrix[row, col].points;
                                    update_points.Add(new Point(row + pkg_current.delete[i].Y, col + pkg_current.delete[i].X));
                                }
                                matrix[row, col].orientation = get_next_Orientation(matrix[row, col].orientation,"left");
                            }
                        }
                        break;
                    default:
                        break;
                }
            }
            return flag;
        }
        /// <summary>
        /// stops all moving packages
        /// </summary>
        /// <returns>one of the moving packages</returns>
        public Package stop_moving_Blocks()
        {
            Package pkg = new Package();
            for (int row = rows - 1; row >= 0; row--)
            {
                for (int col = 0; col < cols; col++)
                {
                    if (matrix[row, col].state == enum_state.moving)
                    {
                        matrix[row, col].state = enum_state.placed;
                        matrix[row, col].center = false;
                        pkg = matrix[row, col];
                    }
                }
            }
            return pkg;
        }
        /// <summary>
        /// draws the full gameboard
        /// </summary>
        public void draw_full_Gameboard()
        {
            draw_Border();
            for (int row = 5; row < rows-1; row++)
            {
                for (int col = 0; col < cols; col++)
                {
                    if (matrix[row, col].type == enum_type.air && matrix[row, col].state == enum_state.placed)
                    {
                        frm_Paketbote.form_Graphics.FillRectangle(frm_Paketbote.brush_white, new Rectangle(col * 40 + top.X, (row - 5) * 40 + top.Y, 40, 40));
                    }
                    else if (matrix[row, col].type != enum_type.air)
                    {
                        frm_Paketbote.form_Graphics.FillRectangle(frm_Paketbote.brush_pkg, new Rectangle(col * 40 + top.X, (row - 5) * 40 + top.Y, 40, 40));
                    }
                    else{}
                }
            }
        }
        /// <summary>
        /// draws only the slots of the gameboard which were updated for better performance
        /// </summary>
        public void draw_partial_Gameboard()
        {
            for (int i = 0; i < update_points.Count; i++)
            {
                if (update_points[i].X > 4)
                {
                    if (matrix[update_points[i].X, update_points[i].Y].type == enum_type.air && matrix[update_points[i].X, update_points[i].Y].state == enum_state.placed)
                    {
                        frm_Paketbote.form_Graphics.FillRectangle(frm_Paketbote.brush_white, new Rectangle(update_points[i].Y * 40 + top.X, (update_points[i].X - 5) * 40 + top.Y, 40, 40));
                    }
                    else if (matrix[update_points[i].X, update_points[i].Y].type != enum_type.air)
                    {
                        frm_Paketbote.form_Graphics.FillRectangle(frm_Paketbote.brush_pkg, new Rectangle(update_points[i].Y * 40 + top.X, (update_points[i].X - 5) * 40 + top.Y, 40, 40));
                    }
                    else{}
                }
            }
        }
    }
}