﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Paketbote
{

    /// <summary>
    /// current state of package
    /// </summary>
    public enum enum_state
    {
        moving,
        placed
    }
    /// <summary>
    /// type of package
    /// </summary>
    public enum enum_type
    {
        air,
        ground,
        T,
        L,
        J,
        Z,
        S,
        I,
        O
    }
    /// <summary>
    /// orientation of package (zero : default)
    /// </summary>
    public enum enum_orientation
    {
        zero,
        one,
        two,
        three
    }

    class Package
    {
        public enum_state state{ get; set; }
        public enum_type type { get; set; }
        public bool center { get; set; }
        public enum_orientation orientation { get; set; }
        public int points { get; set; }

        /// <summary>
        /// creates a new instance of a package
        /// </summary>
        /// <param name="type">type of the package</param>
        /// <param name="state">state of the package</param>
        /// <param name="points">points when package was placed</param>
        /// <param name="center">true if this is the center of the whole package</param>
        public Package(enum_type type, enum_state state, enum_orientation orientation, int points, bool center)
        {
            this.type = type;
            this.state = state;
            this.orientation = orientation;
            this.points = points;
            this.center = center;
        }
        /// <summary>
        /// creates a new empty (air,placed,0,false) package
        /// </summary>
        public Package()
        {
            this.type = enum_type.air;
            this.state = enum_state.placed;
            this.points = 0;
            this.center = false;
        }
        /// <summary>
        /// returns the points for the package type
        /// </summary>
        /// <param name="type">type of package</param>
        /// <returns>the points for the package type</returns>
        public static int get_Points(enum_type type)
        {
            switch (type)
            {
                case enum_type.T:
                    return 2;
                case enum_type.L:
                    return 4;
                case enum_type.J:
                    return 4;
                case enum_type.Z:
                    return 3;
                case enum_type.S:
                    return 3;
                case enum_type.I:
                    return 2;
                case enum_type.O:
                    return 2;
                default:
                    return 0;
            }
        }
    }
}
